#
# This file is part of LiteX.
#
# Copyright (c) 2024 Fin Maaß <f.maass@vogl-electronic.com>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *

from litex.gen import *

import math

from litex.build.io import SDRTristate


def freq_to_div(sys_clk_freq, freq):
    return math.ceil(sys_clk_freq / (4*freq)) - 1

class LiteI2CClkGen(LiteXModule):
    """I2C Clock generator

    The ``LiteI2CClkGen`` class provides a generic I2C clock generator.

    Parameters
    ----------
    pads : Object
        i2C pads description.

    i2c_speed_mode : Signal(2), in
        I2C speed mode.
    
    sys_clk_freq : int
        System clock frequency.

    Attributes
    ----------
    en : Signal(), in
        Clock enable input, output clock will be generated if set to 1, 0 resets the core.

    tx : Signal(), out
        Outputs 1 when the clock is high and the I2C bus is in the transmit state.

    rx : Signal(), out
        Outputs 1 when the clock is low and the I2C bus is in the receive state.

    keep_low : Signal(), in
        Forces the clock to be low, when the clock is disabled.

    suppress : Signal(), in
        Disables the clock output.
    """
    def __init__(self, pads, i2c_speed_mode, sys_clk_freq):
        self.tx         = tx         = Signal()
        self.rx         = rx         = Signal()
        self.en         = en         = Signal()
        self.keep_low   = keep_low   = Signal()
        self.suppress   = suppress   = Signal()
    
        cnt_width = bits_for(freq_to_div(sys_clk_freq, 100000))
    
        div             = Signal(cnt_width)
        cnt             = Signal(cnt_width)
        sub_cnt         = Signal(2)
        clk             = Signal(reset=1)

        self.comb += [
            Case(i2c_speed_mode, {
                0 : div.eq(freq_to_div(sys_clk_freq, 100000)),  #  100 kHz (Standard Mode)
                1 : div.eq(freq_to_div(sys_clk_freq, 400000)),  #  400 kHz (Fast Mode)
                2 : div.eq(freq_to_div(sys_clk_freq, 1000000)), # 1000 kHz (Fast Mode Plus)
            })]

        self.comb += [
            tx.eq(en & (sub_cnt == 0b01) & (cnt == div)),
            rx.eq(en & (sub_cnt == 0b11) & (cnt == div)),
        ]

        self.sync += [
            If(en,
                If(cnt < div,
                    cnt.eq(cnt+1),
                ).Else(
                    cnt.eq(0),
                    clk.eq(sub_cnt[1]),
                    If(sub_cnt < 3,
                        sub_cnt.eq(sub_cnt+1),
                    ).Else(
                        sub_cnt.eq(0),
                    )
                )
            ).Else(
                clk.eq(~keep_low),
                cnt.eq(0),
                sub_cnt.eq(0),
            )
        ]

        self.specials += SDRTristate(
            io = pads.scl,
            o  = Constant(0),       # I2C uses Pull-ups, only drive low.
            oe = ~clk & ~suppress,  # Drive when scl is low and not suppressed.
            i  = Signal(),          # Not used.
        )




#
# This file is part of LiteI2C.
#
# Copyright (c) 2024 Fin Maaß <f.maass@vogl-electronic.com>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *
from migen.genlib.cdc import MultiReg

# Core <-> PHY Layouts -----------------------------------------------------------------------------

"""
Stream layout for LiteI2CCore->PHY connection:
data - data to be transmitted
addr - slave address
len_tx - number of bytes to transmit
len_rx - number of bytes to receive
"""
i2c_core2phy_layout = [
    ("data",     32),
    ("addr",     7),
    ("len_tx",   3),
    ("len_rx",   3),
    ("recover",  1)
]
"""
Stream layout for PHY->LiteI2CCore connection
data - received data
nack - NACK signal
unfinished_tx - another tx transfer is expected
unfinished_rx - another rx transfer is expected
"""
i2c_phy2core_layout = [
    ("data",  32),
    ("nack", 1),
    ("unfinished_tx", 1),
    ("unfinished_rx", 1)
]

# Helpers ------------------------------------------------------------------------------------------

class ResyncReg(Module):
    def __init__(self, src, dst, clock_domain):
        if clock_domain == "sys":
            self.comb += dst.eq(src)
        else:
            self.specials += MultiReg(src, dst, clock_domain)



#
# This file is part of LiteI2C.
#
# Copyright (c) 2024 Fin Maaß <f.maass@vogl-electronic.com>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *

from litex.gen import *

from litex.soc.interconnect import stream
from litex.soc.interconnect.csr import *

from litei2c.common import *
from litei2c.crossbar import LiteI2CCrossbar
from litei2c.core.master import LiteI2CMaster
from litei2c.phy.generic import LiteI2CPHYCore


class LiteI2CCore(Module):
    def __init__(self):
        self.source = stream.Endpoint(i2c_core2phy_layout)
        self.sink   = stream.Endpoint(i2c_phy2core_layout)
        self.active = Signal()


class LiteI2C(LiteXModule):
    """I2C Controller wrapper.

    The ``LiteI2C`` class provides a wrapper that can instantiate ``LiteI2CMaster`` and connect it to the PHY.

    Access to PHY can be shared via crossbar.

    Parameters
    ----------
    sys_clk_freq : int
        Frequency of the system clock.

    phy : Module
        Module or object that contains PHY stream interfaces and a active signal to connect
        the ``LiteI2C`` to. If not provided, it will be created automatically based on the pads.
    
    pads : Object
        I2C pads description.

    clock_domain : str
        Name of LiteI2C clock domain.

    with_master : bool
        Enables register-operated I2C master controller.

    """

    def __init__(self, sys_clk_freq, phy=None, pads=None, clock_domain="sys",
        with_master=True, i2c_master_tx_fifo_depth=1, i2c_master_rx_fifo_depth=1):

        if phy is None:
            if pads is None:
                raise ValueError("Either phy or pads must be provided.")
            self.phy = phy = LiteI2CPHYCore(pads, clock_domain, sys_clk_freq)


        self.crossbar = crossbar = LiteI2CCrossbar(clock_domain)

        self.comb += phy.active.eq(crossbar.active)

        if with_master:
            self.master = master = LiteI2CMaster(
                tx_fifo_depth = i2c_master_tx_fifo_depth,
                rx_fifo_depth = i2c_master_rx_fifo_depth)
            port_master = crossbar.get_port(master.active)
            self.comb += [
                port_master.source.connect(master.sink),
                master.source.connect(port_master.sink),
            ]

        if clock_domain != "sys":
            self.comb += [
                crossbar.tx_cdc.source.connect(phy.sink),
                phy.source.connect(crossbar.rx_cdc.sink),
            ]
        else:
            self.comb += [
                crossbar.master.source.connect(phy.sink),
                phy.source.connect(crossbar.master.sink),
            ]

    def add_i2c_device(self, i2c_device):
        port = self.crossbar.get_port(i2c_device.active)
        self.comb += [
            port.source.connect(i2c_device.sink),
            i2c_device.source.connect(port.sink),
        ]
        



#!/usr/bin/env python3

from setuptools import setup
from setuptools import find_packages

with open("README.md", "r") as fp:
     long_description = fp.read()

setup(
    name                 = "litei2c",
    version              = "2024.12",
    description          = "Small footprint and configurable I2C core",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author               = "LiteI2C Developers",
    url                  = "https://github.com/litex-hub",
    download_url         = "https://github.com/litex-hub/litei2c",
    test_suite           = "test",
    license              = "BSD",
    python_requires      = "~=3.7",
    packages             = find_packages(exclude=("test*", "sim*", "doc*", "examples*")),
    include_package_data = True,
    install_requires=["migen", "litex"],
    extras_require={
        "develop": [
           "meson"
           "pexpect"
           "setuptools"
           "requests"
        ]
    },
    keywords             = "HDL ASIC FPGA hardware design",
    classifiers          = [
        "Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)",
        "Environment :: Console",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
    ],
)
